// bla.c for MacOS   by 川合秀実  ライセンスはKL-01.

// このプログラムは OSECPU-VM rev2 の driver.c のソースコードを大いに参考にしました.
// hikaliumさん, Livaさん, rynaさんに深く感謝いたします.
// またロシアンブルー(古川菜摘)さんにもたくさんアドバイスをもらっています.

#include "bla.h"
#include <stdlib.h>
#include <mach/mach.h>

#define bla_TABLESIZ	16

static bla_Window *bla_table_w[bla_TABLESIZ];

static void putKeybuf(bla_Window *w, int c)
{
	if (((w->kbw + 1) & (bla_KEYBUFSIZ - 1)) != w->kbr) {
		w->keybuf[w->kbw] = c;
		w->kbw = (w->kbw + 1) & (bla_KEYBUFSIZ - 1);
	}
	return;
}

// level0

@interface BlaView : NSView {
	bla_Window *_w;
	CGContextRef _context;
}

- (id) initWithBla: (bla_Window *) w;
- (void) drawRect: (NSRect) rect; // OSがこれを呼び出す.

@end

@implementation BlaView

- (id) initWithBla: (bla_Window *) w
{
	self = [super initWithFrame: NSMakeRect(0, 0, w->xsiz, w->ysiz)];
	if (self) {
		_w = w;
	}
	return self;
}

- (void) drawRect: (NSRect) rect
{
	CGColorSpaceRef colorSpace = CGColorSpaceCreateDeviceRGB();
	_context = CGBitmapContextCreate(_w->buf, _w->xsiz, _w->ysiz, 8, 4 * _w->xsiz, colorSpace, (kCGBitmapByteOrder32Little | kCGImageAlphaNoneSkipFirst));
	CGImageRef image = CGBitmapContextCreateImage(_context);
	CGContextRef currentContext = (CGContextRef)[[NSGraphicsContext currentContext] graphicsPort];
	CGContextDrawImage(currentContext, NSRectToCGRect(rect), image);
	CGColorSpaceRelease(colorSpace);
	CGImageRelease(image);
	return;
}

// キー入力．
// このメソッドが呼ばれるためには，
// このビューがWindowのFirst Responderで無ければならない.
- (void) keyDown: (NSEvent *) theEvent
{
	if ([[theEvent characters] length] <= 0) return;

	// ここでcはキーコードなどでは無く，生成された文字．
	// そのため，OSのキーマップに依存する可能性がある．

	unichar c = [[theEvent characters] characterAtIndex:0];
	NSUInteger flags = [NSEvent modifierFlags];

	int i = c, j = 0;
 
	if (c == '\r')	i = bla_KEY_ENTER;
	if (c == '\e')	i = bla_KEY_ESC;
	if (c == 127)	i = bla_KEY_BACKSPACE;
	if (c == '\t')	i = bla_KEY_TAB;

	if (c == NSPageUpFunctionKey)		i = bla_KEY_PAGEUP;
	if (c == NSPageDownFunctionKey)		i = bla_KEY_PAGEDWN;
	if (c == NSEndFunctionKey)			i = bla_KEY_END;
	if (c == NSHomeFunctionKey)			i = bla_KEY_HOME;
	if (c == NSLeftArrowFunctionKey)	i = bla_KEY_LEFT;
	if (c == NSRightArrowFunctionKey)	i = bla_KEY_RIGHT;
	if (c == NSUpArrowFunctionKey)		i = bla_KEY_UP;
	if (c == NSDownArrowFunctionKey)	i = bla_KEY_DOWN;
	if (c == NSInsertFunctionKey)		i = bla_KEY_INS; // Not on most Macintosh keyboards.
	if (c == NSDeleteFunctionKey)		i = bla_KEY_DEL;

	// Cocoaフレームワークでは，修飾キーの右左を区別するやつが見当たらなかったです.
	if ((flags & NSEventModifierFlagShift) != 0)	i |= 1 << 16;
	if ((flags & NSEventModifierFlagControl) != 0)	j |= 1 << 17;
	if ((flags & NSEventModifierFlagOption) != 0)	j |= 1 << 18;
	if ((flags & NSEventModifierFlagCapsLock) != 0) i |= 1 << 23;
	putKeybuf(_w, i | j);
	return;
}

@end

@interface BlaWinDel : NSObject<NSWindowDelegate> {
	bla_Window *_w;
}

- (id) initWithBla: (bla_Window *) w;
- (BOOL) windowShouldClose: (id) sender; // OSがこれを呼び出す.

@end

@implementation BlaWinDel

- (id) initWithBla: (bla_Window *) w
{
	_w = w;
	return self;
}

- (BOOL) windowShouldClose: (id) sender
{
	_w->phase = 2;
	return YES;
}

@end

bla_Window *bla_openWin(int x, int y, const char *t)
{
	int i, j;
	static char init = 0;
	if (init == 0) {
		ProcessSerialNumber psn = {0, kCurrentProcess};
		// ForegroundApplicationになれるように設定(こうしないとキーイベントが取得できない)
		TransformProcessType(&psn, kProcessTransformToForegroundApplication);
		init = 1;
	}
	for (i = 0; i < bla_TABLESIZ; i++) {
		if (bla_table_w[i] == 0) break;
	}
	if (i >= bla_TABLESIZ) return 0;

	bla_Window *w = malloc(sizeof (bla_Window));
	bla_table_w[i] = w;
	w->table_idx = i;
	w->title[0] = '\0';
	if (t != 0) {
		strncpy(w->title, t, 63);
		w->title[63] = '\0';
	}
	if (w->title[0] == '\0') {
		w->title[0] = ' ';
		w->title[1] = '\0';
	}
	w->kbw = w->kbr = 0;
	w->phase = 1;
	w->xsiz = x;
	w->ysiz = y;
	w->buf = malloc(x * y * sizeof (int));
	w->lastFlush = clock();
	w->mode = bla_MODE_SET;
#if 0
	w->cx = pwtp->cy = 0;
	w->col = 0xffffff;
	w->bcol = 0;
	w->gets_ph = 0;
#endif
	bla_fillRect(w, x, y, 0, 0, 0);

	NSWindow *win = [[NSWindow alloc] initWithContentRect: NSMakeRect(0, 0, x, y) styleMask: NSWindowStyleMaskTitled | NSWindowStyleMaskMiniaturizable | NSWindowStyleMaskClosable backing: NSBackingStoreBuffered defer: NO];
	w->win = win;
	w->tit = [[NSString alloc] initWithUTF8String: w->title];
	[win setTitle: w->tit];
	[win center];
	[win makeKeyAndOrderFront: nil];
	[win setReleasedWhenClosed: YES];
	w->del = [[BlaWinDel alloc] initWithBla: w];
	win.delegate = w->del;
	BlaView *view = [[BlaView alloc] initWithBla: w];
	w->view = view;
	[win.contentView addSubview: view];
	[win makeFirstResponder: view];
	return w;
}

void bla_flushAll(bla_Window *w)
{
	if (w->phase == 1) {
		[w->view setNeedsDisplayInRect: NSMakeRect(0, 0, w->xsiz, w->ysiz)];
		w->lastFlush = clock();
//		bla_wait(1); // これをやらないと、flushが成功しない. // マルチスレッド化に伴い、不要になった.
	}
	return;
}

void bla_wait(int msec)
{
//	[[NSRunLoop currentRunLoop] runUntilDate:[NSDate dateWithTimeIntervalSinceNow: 0.001 * msec]];
	[NSThread sleepForTimeInterval: 0.001 * msec];
	return;
}

void bla_closeWin(bla_Window *w)
{
	if (w->phase == 1) {
		[w->win close];
		while (w->phase == 1)
			bla_wait(128);
	}
	[w->win release];
	[w->view release];
	[w->del release];
	[w->tit release];
	free(w->buf);
	bla_table_w[w->table_idx] = 0;
	free(w);
	return;
}

void bla_exit(int c) { exit(c); }

@interface Main : NSObject<NSWindowDelegate> {
	int argc;
	char **argv;
}

- (void) runApp;
- (void) createThread: (int) _argc args: (char **) _argv;

@end

@implementation Main

- (void) runApp
{
	NSAutoreleasePool* pool = [[NSAutoreleasePool alloc] init];
	bla_main(argc, argv);
	[NSApp terminate: self];
	[pool release];
}

- (void) createThread: (int) _argc args: (char **) _argv
{
	argc = _argc;
	argv = _argv;
	NSThread *thread = [[[NSThread alloc] initWithTarget: self selector: @selector(runApp) object: nil] autorelease];
	[thread start];
}

@end

int main(int argc, char **argv)
{
	[NSApplication sharedApplication];

	// note(hikalium):
	// As of Snow Leopard, programs without application bundles and Info.plist files
	// don't get a menubar and can't be brought to the front
	// unless the presentation option is changed.
	// See: https://www.cocoawithlove.com/2010/09/minimalist-cocoa-programming.html
	[NSApp setActivationPolicy:NSApplicationActivationPolicyRegular];

	id thread_blamain = [[Main alloc] init];
	[thread_blamain createThread: argc args: argv];
	[NSApp run];
	return 0;
}

// level1

void bla_end()
{
	int i, j;
	for (i = 0; i < bla_TABLESIZ; i++) {
		bla_Window *w = bla_table_w[i];
		if (w == 0) continue;
		bla_flushAll(w);
	}
	for (;;) {
		for (i = 0; i < bla_TABLESIZ; i++) {
			bla_Window *w = bla_table_w[i];
			if (w == 0) continue;
			if (w->phase == 2)
				bla_closeWin(w);
		}
		j = 0;
		for (i = 0; i < bla_TABLESIZ; i++) {
			bla_Window *w = bla_table_w[i];
			if (w == 0) continue;
			j++;
		}
		if (j == 0) break;
		bla_wait(128);
	}
	bla_exit(0);
}

void bla_leapFlushAll(bla_Window *w, unsigned int t)
{
	if (t == 0 || (double) (clock() - w->lastFlush) * (1000.0 / (double) CLOCKS_PER_SEC) >= (double) t)
		bla_flushAll(w);
	return;
}

int bla_inkey(bla_Window *w)
{
	int i;
	if (w->kbr == w->kbw) return 0;
	i = w->keybuf[w->kbr];
	w->kbr = (w->kbr + 1) & (bla_KEYBUFSIZ - 1);
	return i;
}

int bla_inkey0(bla_Window *w)
{
	if (w->kbr == w->kbw) return 0;
	return w->keybuf[w->kbr];
}

static unsigned char fontdata[] = {
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x00, 0x00, 0x10, 0x10, 0x00, 0x00,
	0x28, 0x28, 0x28, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x44, 0x44, 0x44, 0xfe, 0x44, 0x44, 0x44, 0x44, 0x44, 0xfe, 0x44, 0x44, 0x44, 0x00, 0x00,
	0x10, 0x3a, 0x56, 0x92, 0x92, 0x90, 0x50, 0x38, 0x14, 0x12, 0x92, 0x92, 0xd4, 0xb8, 0x10, 0x10,
	0x62, 0x92, 0x94, 0x94, 0x68, 0x08, 0x10, 0x10, 0x20, 0x2c, 0x52, 0x52, 0x92, 0x8c, 0x00, 0x00,
	0x00, 0x70, 0x88, 0x88, 0x88, 0x90, 0x60, 0x47, 0xa2, 0x92, 0x8a, 0x84, 0x46, 0x39, 0x00, 0x00,
	0x04, 0x08, 0x10, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x02, 0x04, 0x08, 0x08, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x08, 0x08, 0x04, 0x02, 0x00,
	0x80, 0x40, 0x20, 0x20, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x20, 0x20, 0x40, 0x80, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x10, 0x92, 0x54, 0x38, 0x54, 0x92, 0x10, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x10, 0x10, 0x10, 0xfe, 0x10, 0x10, 0x10, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x18, 0x18, 0x08, 0x08, 0x10,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xfe, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x18, 0x18, 0x00, 0x00,
	0x02, 0x02, 0x04, 0x04, 0x08, 0x08, 0x08, 0x10, 0x10, 0x20, 0x20, 0x40, 0x40, 0x40, 0x80, 0x80,
	0x00, 0x18, 0x24, 0x24, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x24, 0x24, 0x18, 0x00, 0x00,
	0x00, 0x08, 0x18, 0x28, 0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 0x3e, 0x00, 0x00,
	0x00, 0x18, 0x24, 0x42, 0x42, 0x02, 0x04, 0x08, 0x10, 0x20, 0x20, 0x40, 0x40, 0x7e, 0x00, 0x00,
	0x00, 0x18, 0x24, 0x42, 0x02, 0x02, 0x04, 0x18, 0x04, 0x02, 0x02, 0x42, 0x24, 0x18, 0x00, 0x00,
	0x00, 0x0c, 0x0c, 0x0c, 0x14, 0x14, 0x14, 0x24, 0x24, 0x44, 0x7e, 0x04, 0x04, 0x1e, 0x00, 0x00,
	0x00, 0x7c, 0x40, 0x40, 0x40, 0x58, 0x64, 0x02, 0x02, 0x02, 0x02, 0x42, 0x24, 0x18, 0x00, 0x00,
	0x00, 0x18, 0x24, 0x42, 0x40, 0x58, 0x64, 0x42, 0x42, 0x42, 0x42, 0x42, 0x24, 0x18, 0x00, 0x00,
	0x00, 0x7e, 0x42, 0x42, 0x04, 0x04, 0x08, 0x08, 0x08, 0x10, 0x10, 0x10, 0x10, 0x38, 0x00, 0x00,
	0x00, 0x18, 0x24, 0x42, 0x42, 0x42, 0x24, 0x18, 0x24, 0x42, 0x42, 0x42, 0x24, 0x18, 0x00, 0x00,
	0x00, 0x18, 0x24, 0x42, 0x42, 0x42, 0x42, 0x42, 0x26, 0x1a, 0x02, 0x42, 0x24, 0x18, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x18, 0x18, 0x00, 0x00, 0x00, 0x00, 0x00, 0x18, 0x18, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x18, 0x18, 0x00, 0x00, 0x00, 0x00, 0x18, 0x18, 0x08, 0x08, 0x10,
	0x00, 0x02, 0x04, 0x08, 0x10, 0x20, 0x40, 0x80, 0x80, 0x40, 0x20, 0x10, 0x08, 0x04, 0x02, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xfe, 0x00, 0x00, 0xfe, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x80, 0x40, 0x20, 0x10, 0x08, 0x04, 0x02, 0x02, 0x04, 0x08, 0x10, 0x20, 0x40, 0x80, 0x00,
	0x00, 0x38, 0x44, 0x82, 0x82, 0x82, 0x04, 0x08, 0x10, 0x10, 0x00, 0x00, 0x18, 0x18, 0x00, 0x00,
	0x00, 0x38, 0x44, 0x82, 0x9a, 0xaa, 0xaa, 0xaa, 0xaa, 0xaa, 0x9c, 0x80, 0x46, 0x38, 0x00, 0x00,
	0x00, 0x18, 0x18, 0x18, 0x18, 0x24, 0x24, 0x24, 0x24, 0x7e, 0x42, 0x42, 0x42, 0xe7, 0x00, 0x00,
	0x00, 0xf0, 0x48, 0x44, 0x44, 0x44, 0x48, 0x78, 0x44, 0x42, 0x42, 0x42, 0x44, 0xf8, 0x00, 0x00,
	0x00, 0x3a, 0x46, 0x42, 0x82, 0x80, 0x80, 0x80, 0x80, 0x80, 0x82, 0x42, 0x44, 0x38, 0x00, 0x00,
	0x00, 0xf8, 0x44, 0x44, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x44, 0x44, 0xf8, 0x00, 0x00,
	0x00, 0xfe, 0x42, 0x42, 0x40, 0x40, 0x44, 0x7c, 0x44, 0x40, 0x40, 0x42, 0x42, 0xfe, 0x00, 0x00,
	0x00, 0xfe, 0x42, 0x42, 0x40, 0x40, 0x44, 0x7c, 0x44, 0x44, 0x40, 0x40, 0x40, 0xf0, 0x00, 0x00,
	0x00, 0x3a, 0x46, 0x42, 0x82, 0x80, 0x80, 0x9e, 0x82, 0x82, 0x82, 0x42, 0x46, 0x38, 0x00, 0x00,
	0x00, 0xe7, 0x42, 0x42, 0x42, 0x42, 0x42, 0x7e, 0x42, 0x42, 0x42, 0x42, 0x42, 0xe7, 0x00, 0x00,
	0x00, 0x7c, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x7c, 0x00, 0x00,
	0x00, 0x1f, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x84, 0x48, 0x30, 0x00,
	0x00, 0xe7, 0x42, 0x44, 0x48, 0x50, 0x50, 0x60, 0x50, 0x50, 0x48, 0x44, 0x42, 0xe7, 0x00, 0x00,
	0x00, 0xf0, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x42, 0x42, 0xfe, 0x00, 0x00,
	0x00, 0xc3, 0x42, 0x66, 0x66, 0x66, 0x5a, 0x5a, 0x5a, 0x42, 0x42, 0x42, 0x42, 0xe7, 0x00, 0x00,
	0x00, 0xc7, 0x42, 0x62, 0x62, 0x52, 0x52, 0x52, 0x4a, 0x4a, 0x4a, 0x46, 0x46, 0xe2, 0x00, 0x00,
	0x00, 0x38, 0x44, 0x82, 0x82, 0x82, 0x82, 0x82, 0x82, 0x82, 0x82, 0x82, 0x44, 0x38, 0x00, 0x00,
	0x00, 0xf8, 0x44, 0x42, 0x42, 0x42, 0x44, 0x78, 0x40, 0x40, 0x40, 0x40, 0x40, 0xf0, 0x00, 0x00,
	0x00, 0x38, 0x44, 0x82, 0x82, 0x82, 0x82, 0x82, 0x82, 0x82, 0x92, 0x8a, 0x44, 0x3a, 0x00, 0x00,
	0x00, 0xfc, 0x42, 0x42, 0x42, 0x42, 0x7c, 0x44, 0x42, 0x42, 0x42, 0x42, 0x42, 0xe7, 0x00, 0x00,
	0x00, 0x3a, 0x46, 0x82, 0x82, 0x80, 0x40, 0x38, 0x04, 0x02, 0x82, 0x82, 0xc4, 0xb8, 0x00, 0x00,
	0x00, 0xfe, 0x92, 0x92, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x7c, 0x00, 0x00,
	0x00, 0xe7, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x24, 0x3c, 0x00, 0x00,
	0x00, 0xe7, 0x42, 0x42, 0x42, 0x42, 0x24, 0x24, 0x24, 0x24, 0x18, 0x18, 0x18, 0x18, 0x00, 0x00,
	0x00, 0xe7, 0x42, 0x42, 0x42, 0x5a, 0x5a, 0x5a, 0x5a, 0x24, 0x24, 0x24, 0x24, 0x24, 0x00, 0x00,
	0x00, 0xe7, 0x42, 0x42, 0x24, 0x24, 0x24, 0x18, 0x24, 0x24, 0x24, 0x42, 0x42, 0xe7, 0x00, 0x00,
	0x00, 0xee, 0x44, 0x44, 0x44, 0x28, 0x28, 0x28, 0x10, 0x10, 0x10, 0x10, 0x10, 0x7c, 0x00, 0x00,
	0x00, 0xfe, 0x84, 0x84, 0x08, 0x08, 0x10, 0x10, 0x20, 0x20, 0x40, 0x42, 0x82, 0xfe, 0x00, 0x00,
	0x00, 0x3e, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x3e, 0x00,
	0x80, 0x80, 0x40, 0x40, 0x20, 0x20, 0x20, 0x10, 0x10, 0x08, 0x08, 0x04, 0x04, 0x04, 0x02, 0x02,
	0x00, 0x7c, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x7c, 0x00,
	0x00, 0x10, 0x28, 0x44, 0x82, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xfe, 0x00,
	0x10, 0x08, 0x04, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x70, 0x08, 0x04, 0x3c, 0x44, 0x84, 0x84, 0x8c, 0x76, 0x00, 0x00,
	0xc0, 0x40, 0x40, 0x40, 0x40, 0x58, 0x64, 0x42, 0x42, 0x42, 0x42, 0x42, 0x64, 0x58, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x30, 0x4c, 0x84, 0x84, 0x80, 0x80, 0x82, 0x44, 0x38, 0x00, 0x00,
	0x0c, 0x04, 0x04, 0x04, 0x04, 0x34, 0x4c, 0x84, 0x84, 0x84, 0x84, 0x84, 0x4c, 0x36, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x38, 0x44, 0x82, 0x82, 0xfc, 0x80, 0x82, 0x42, 0x3c, 0x00, 0x00,
	0x0e, 0x10, 0x10, 0x10, 0x10, 0x7c, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x7c, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x36, 0x4c, 0x84, 0x84, 0x84, 0x84, 0x4c, 0x34, 0x04, 0x04, 0x78,
	0xc0, 0x40, 0x40, 0x40, 0x40, 0x58, 0x64, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0xe3, 0x00, 0x00,
	0x00, 0x10, 0x10, 0x00, 0x00, 0x30, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x38, 0x00, 0x00,
	0x00, 0x04, 0x04, 0x00, 0x00, 0x0c, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x08, 0x08, 0x30,
	0xc0, 0x40, 0x40, 0x40, 0x40, 0x4e, 0x44, 0x48, 0x50, 0x60, 0x50, 0x48, 0x44, 0xe6, 0x00, 0x00,
	0x30, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x38, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0xf6, 0x49, 0x49, 0x49, 0x49, 0x49, 0x49, 0x49, 0xdb, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0xd8, 0x64, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0xe3, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x38, 0x44, 0x82, 0x82, 0x82, 0x82, 0x82, 0x44, 0x38, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0xd8, 0x64, 0x42, 0x42, 0x42, 0x42, 0x42, 0x64, 0x58, 0x40, 0xe0,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x34, 0x4c, 0x84, 0x84, 0x84, 0x84, 0x84, 0x4c, 0x34, 0x04, 0x0e,
	0x00, 0x00, 0x00, 0x00, 0x00, 0xdc, 0x62, 0x42, 0x40, 0x40, 0x40, 0x40, 0x40, 0xe0, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x7a, 0x86, 0x82, 0xc0, 0x38, 0x06, 0x82, 0xc2, 0xbc, 0x00, 0x00,
	0x00, 0x00, 0x10, 0x10, 0x10, 0x7c, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x0e, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0xc6, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x46, 0x3b, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0xe7, 0x42, 0x42, 0x42, 0x24, 0x24, 0x24, 0x18, 0x18, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0xe7, 0x42, 0x42, 0x5a, 0x5a, 0x5a, 0x24, 0x24, 0x24, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0xc6, 0x44, 0x28, 0x28, 0x10, 0x28, 0x28, 0x44, 0xc6, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0xe7, 0x42, 0x42, 0x24, 0x24, 0x24, 0x18, 0x18, 0x10, 0x10, 0x60,
	0x00, 0x00, 0x00, 0x00, 0x00, 0xfe, 0x82, 0x84, 0x08, 0x10, 0x20, 0x42, 0x82, 0xfe, 0x00, 0x00,
	0x00, 0x06, 0x08, 0x10, 0x10, 0x10, 0x10, 0x60, 0x10, 0x10, 0x10, 0x10, 0x08, 0x06, 0x00, 0x00,
	0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10,
	0x00, 0x60, 0x10, 0x08, 0x08, 0x08, 0x08, 0x06, 0x08, 0x08, 0x08, 0x08, 0x10, 0x60, 0x00, 0x00,
	0x00, 0x72, 0x8c, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x10, 0x28, 0x44, 0x82, 0xfe, 0x82, 0xfe, 0x00, 0x00, 0x00, 0x00, 0x00
};

void bla_setPixEx(bla_Window *w, int x, int y, int c)
{
	if (c >= 0 && 0 <= x && x < w->xsiz && 0 <= y && y < w->ysiz) {
		if (w->mode == bla_MODE_SET) w->buf[y * w->xsiz + x] =  c;
		if (w->mode == bla_MODE_OR)  w->buf[y * w->xsiz + x] |= c;
		if (w->mode == bla_MODE_AND) w->buf[y * w->xsiz + x] &= c;
		if (w->mode == bla_MODE_XOR) w->buf[y * w->xsiz + x] ^= c;
	}
	return;
}

void bla_fillRect(bla_Window *w, int sx, int sy, int x0, int y0, int c)
{
	int i, j, *b = w->buf + x0 + y0 * w->xsiz;
	for (j = 0; j < sy; j++) {
		for (i = 0; i < sx; i++)
			b[i] = c;
		b += w->xsiz;
	}
	return;
}

void bla_fillRectEx(bla_Window *w, int sx, int sy, int x0, int y0, int c)
{
	int i, j;
	for (j = 0; j < sy; j++) {
		for (i = 0; i < sx; i++) {
			bla_setPixEx(w, x0 + i, y0 + j, c);
		}
	}
	return;
}

void bla_drawRect(bla_Window *w, int sx, int sy, int x0, int y0, int c)
{
	if (sx <= 1 || sy <= 1)
		bla_fillRect(w, sx, sy, x0, y0, c);
	else {
		bla_fillRect(w, sx, 1,      x0,          y0,          c);
		bla_fillRect(w, sx, 1,      x0,          y0 + sy - 1, c);
		bla_fillRect(w, 1,  sy - 2, x0,          y0 + 1,      c);
		bla_fillRect(w, 1,  sy - 2, x0 + sx - 1, y0 + 1,      c);
	}
	return;
}

void bla_drawRectEx(bla_Window *w, int sx, int sy, int x0, int y0, int c)
{
	if (sx <= 1 || sy <= 1)
		bla_fillRectEx(w, sx, sy, x0, y0, c);
	else {
		bla_fillRectEx(w, sx, 1,      x0,          y0,          c);
		bla_fillRectEx(w, sx, 1,      x0,          y0 + sy - 1, c);
		bla_fillRectEx(w, 1,  sy - 2, x0,          y0 + 1,      c);
		bla_fillRectEx(w, 1,  sy - 2, x0 + sx - 1, y0 + 1,      c);
	}
	return;
}

void bla_drawStr(bla_Window *w, int x, int y, int c, int b, const char *s)
{
	int i, j, k, ch, pt;
	for (i = 0; s[i] != '\0'; i++) {
		ch = s[i];
		if (ch < ' ' || ch > 0x7f) ch = ' ';
		ch = (ch - ' ') * 16;
		for (j = 0; j < 16; j++) {
			pt = fontdata[ch + j];
			for (k = 0; k < 8; k++) {
				if ((pt & 0x80) != 0) {
					bla_setPix(w, x + k, y + j, c);
				} else {
					bla_setPix(w, x + k, y + j, b);
				}
				pt <<= 1;
			}
		}
		x += 8;
	}
	return;
}

void bla_drawStrEx(bla_Window *w, int x, int y, int c, int b, const char *s)
{
	int i, j, k, ch, pt;
	for (i = 0; s[i] != '\0'; i++) {
		ch = s[i];
		if (ch < ' ' || ch > 0x7f) ch = ' ';
		ch = (ch - ' ') * 16;
		for (j = 0; j < 16; j++) {
			pt = fontdata[ch + j];
			for (k = 0; k < 8; k++) {
				if ((pt & 0x80) != 0) {
					bla_setPixEx(w, x + k, y + j, c);
				} else {
					bla_setPixEx(w, x + k, y + j, b);
				}
				pt <<= 1;
			}
		}
		x += 8;
	}
	return;
}

void bla_drawLine(bla_Window *w, int x0, int y0, int x1, int y1, int c)
{
	int x, y, dx, dy, sx, sy, err, e2;
	dx = x1 - x0;
	dy = y1 - y0;
	sx = sy = 1;
	if (dx < 0) { dx *= -1; sx = -1; }
	if (dy < 0) { dy *= -1; sy = -1; }
	err = dx - dy;
	x = x0;
	y = y0;
	for (;;) {
		bla_setPix(w, x, y, c);
		if (x == x1 && y == y1) break;
		e2 = err * 2;
		if (e2 > - dy) {
			err -= dy;
			x += sx;
		}
		if (e2 <   dx) {
			err += dx;
			y += sy;
		}
	}
	return;
}

void bla_drawLineEx(bla_Window *w, int x0, int y0, int x1, int y1, int c)
{
	int x, y, dx, dy, sx, sy, err, e2;
	dx = x1 - x0;
	dy = y1 - y0;
	sx = sy = 1;
	if (dx < 0) { dx *= -1; sx = -1; }
	if (dy < 0) { dy *= -1; sy = -1; }
	err = dx - dy;
	x = x0;
	y = y0;
	for (;;) {
		bla_setPixEx(w, x, y, c);
		if (x == x1 && y == y1) break;
		e2 = err * 2;
		if (e2 > - dy) {
			err -= dy;
			x += sx;
		}
		if (e2 <   dx) {
			err += dx;
			y += sy;
		}
	}
	return;
}

void bla_fillOval(bla_Window *w, int sx, int sy, int x0, int y0, int c)
{
	double dcx, dcy, dcxy, dtx, dty;
	int x, y;
	dcx = 0.5 * (sx - 1);
	dcy = 0.5 * (sy - 1);
	dcxy = (dcx + 0.5) * (dcy + 0.5) - 0.1;
	dcxy *= dcxy;
	for (y = 0; y < sy; y++) {
		dty = (y - dcy) * dcx;
		for (x = 0; x < sx; x++) {
			dtx = (x - dcx) * dcy;
			if (dtx * dtx + dty * dty > dcxy) continue;
				bla_setPix(w, x + x0, y + y0, c);
		}
	}
	return;
}

void bla_fillOvalEx(bla_Window *w, int sx, int sy, int x0, int y0, int c)
{
	double dcx, dcy, dcxy, dtx, dty;
	int x, y;
	dcx = 0.5 * (sx - 1);
	dcy = 0.5 * (sy - 1);
	dcxy = (dcx + 0.5) * (dcy + 0.5) - 0.1;
	dcxy *= dcxy;
	for (y = 0; y < sy; y++) {
		dty = (y - dcy) * dcx;
		for (x = 0; x < sx; x++) {
			dtx = (x - dcx) * dcy;
			if (dtx * dtx + dty * dty > dcxy) continue;
				bla_setPixEx(w, x + x0, y + y0, c);
		}
	}
	return;
}

#define DRAWOVALPARAM	1

void bla_drawOvalEx(bla_Window *w, int sx, int sy, int x0, int y0, int c)
{
	double dcx, dcy, dcxy, dtx, dty, dcx1, dcy1, dcxy1, dtx1, dty1;
	int x, y;
	dcx = 0.5 * (sx - 1);
	dcy = 0.5 * (sy - 1);
	dcxy = (dcx + 0.5) * (dcy + 0.5) - 0.1;
	dcxy *= dcxy;
	dcx1 = 0.5 * (sx - (1 + DRAWOVALPARAM * 2));
	dcy1 = 0.5 * (sy - (1 + DRAWOVALPARAM * 2));
	dcxy1 = (dcx1 + 0.5) * (dcy1 + 0.5) - 0.1;
	dcxy1 *= dcxy1;
	for (y = 0; y < sy; y++) {
		dty  = (y - dcy) * dcx;
		dty1 = (y - dcy) * dcx1;
		for (x = 0; x < sx; x++) {
			dtx = (x - dcx) * dcy;
			dtx1 = (x - dcx) * dcy1;
			if (dtx * dtx + dty * dty > dcxy) continue;
			if (DRAWOVALPARAM <= x && x < sx - DRAWOVALPARAM && DRAWOVALPARAM <= y && y < sy - DRAWOVALPARAM) {
				if (dtx1 * dtx1 + dty1 * dty1 < dcxy1) continue;
			}
			bla_setPixEx(w, x + x0, y + y0, c);
		}
	}
	return;
}

typedef struct bla_Point_ {
	int x, y;
} bla_Point;

void bla_fillEx(bla_Window *w, int x, int y, int c)
{
	int b, sp, sx = w->xsiz, sy = w->ysiz, u, v, *q = w->buf;
	bla_Point *p;
	if (x < 0 || x >= sx || y < 0 || y >= sy || c < 0) return;
	b = q[x + y * sx];
	if (b == c) return;
	p = malloc(sx * sy * sizeof (bla_Point));
	sp = 0; p[sp].x = x; p[sp].y = y; q[x + y * sx] = c;
	do {
		x = p[sp].x;
		y = p[sp].y;
		sp--;
		u = x;     v = y + 1; if (v < sy && q[u + v * sx] == b) { sp++; p[sp].x = u; p[sp].y = v; q[u + v * sx] = c; }
		u = x;     v = y - 1; if (v >= 0 && q[u + v * sx] == b) { sp++; p[sp].x = u; p[sp].y = v; q[u + v * sx] = c; }
		u = x + 1; v = y;     if (u < sx && q[u + v * sx] == b) { sp++; p[sp].x = u; p[sp].y = v; q[u + v * sx] = c; }
		u = x - 1; v = y;     if (u >= 0 && q[u + v * sx] == b) { sp++; p[sp].x = u; p[sp].y = v; q[u + v * sx] = c; }
	} while (sp >= 0);
	free(p);
	return;
}

void bla_waitEx(int msec)
{
	do {
		int i, j = 0;
		for (i = 0; i < bla_TABLESIZ; i++) {
			bla_Window *w = bla_table_w[i];
			if (w == 0) continue;
			if (w->phase == 1) j++;
		}
		if (j == 0) bla_end();
		if (msec > 0) {
			i = msec;
			if (i >= 256)
				i = 128;
			bla_wait(i);
			msec -= i;
		}
	} while (msec > 0);
	return;
}

