#include <stdio.h>
#include <stdlib.h>

static int get32b(const unsigned char *p)
{
	return p[0] | p[1] << 8 | p[2] << 16 | p[3] << 24;
}

int main(int argc, const char **argv)
{
	FILE *fp;
	int sz, i, j, k, l, m, n, secs, optsz, ent, bas, imgsz, subsys, stkRsv, stkCmt, hepRsv, hepCmt, ddsz, idat;
	unsigned char *b = malloc(16 * 1024 * 1024), *im;
	if (argc < 2) {
		fprintf(stderr, "usage>peinfo file\n");
		return 1;
	}
	fp = fopen(argv[1], "rb");
	if (fp == 0) {
		fprintf(stderr, "fopen error\n");
		return 1;
	}
	sz = fread(b, 1, 16 * 1024 * 1024, fp);
	fclose(fp);

	// DOS-MZ
	if (b[0] == 0x4d && b[1] == 0x5a)
		puts("signature 'MZ' found");
	else {
err:
		puts("unknown file format");
		return 0;
	}
	i = get32b(&b[0x3c]);

	// Win-PE
	if (get32b(&b[i + 0]) == 0x00004550)
		puts("signature 'PE..' found");
	else
		goto err;
	i += 4;

	// ImageFileHeader
	if (b[i + 0] == 0x4c && b[i + 1] == 0x01)
		puts("machine: x86(32bit)");
	else
		goto err;
	secs = b[i + 2] | b[i + 3] << 8;
	printf("sections: %d\n", secs);
	optsz = b[i + 16] | b[i + 17] << 8;
	printf("imageOptionalHeader size: 0x%x\n", optsz);
	j = b[i + 18] | b[i + 19] << 8;
	printf("characteristics: 0x%04x\n", j);
	if ((j & 0x2103) == 0x0103)
		;
	else
		goto err;
	i += 20;

	// ImageOptionalHeader
	if (b[i + 0] == 0x0b && b[i + 1] == 0x01)
		printf("type: PE32\n");
	else
		goto err;
	ent = get32b(&b[i + 16]);
	bas = get32b(&b[i + 28]);
	printf("entry: 0x%x\n", ent + bas);
	printf("base:  0x%x\n", bas);
	j = get32b(&b[i + 32]);
	printf("sectionAlign: %d [ref]\n", j);
	j = get32b(&b[i + 36]);
	printf("fileAlign: %d [ref]\n", j);
	imgsz = get32b(&b[i + 56]);
	printf("sizeOfImage: 0x%x\n", imgsz);
	j = get32b(&b[i + 60]);
	printf("sizeOfHeaders: 0x%x [ref]\n", j);
	subsys = b[i + 68] | b[i + 69] << 8;
	printf("subsystem: 0x%x\n", subsys);
	stkRsv = get32b(&b[i + 72]);
	stkCmt = get32b(&b[i + 76]);
	hepRsv = get32b(&b[i + 80]);
	hepCmt = get32b(&b[i + 84]);
	printf("stackReserve: 0x%x\n", stkRsv);
	printf("stackCommit: 0x%x\n", stkCmt);
	printf("heapReserve: 0x%x\n", hepRsv);
	printf("heapCommit: 0x%x\n", hepCmt);
	ddsz = get32b(&b[i + 92]);
	printf("numberOfRvaAndSizes: %d\n", ddsz);
	idat = get32b(&b[i + 96 + 8*1 + 0]);
	j = get32b(&b[i + 96 + 8*1 + 4]);
	printf("idataBase: 0x%x\n", idat + bas);
	printf("idataSize: 0x%x [ref]\n", j);
	i += optsz;

	// SectionHeader
	for (j = 0; j < secs; j++) {
		printf("\n");
		printf("sec[%d].name:    %.8s\n", j, &b[i + j * 40 +  0]);
		printf("sec[%d].size:    0x%x\n", j, get32b(&b[i + j * 40 +  8]));
		printf("sec[%d].addr:    0x%x\n", j, get32b(&b[i + j * 40 + 12]));
		printf("sec[%d].fileSiz: 0x%x\n", j, get32b(&b[i + j * 40 + 16]));
		printf("sec[%d].fileOfs: 0x%x\n", j, get32b(&b[i + j * 40 + 20]));
		k = get32b(&b[i + j * 40 + 36]);
		printf("sec[%d].charact: 0x%x\n", j, k);
		if ((k & 0x00000020) != 0) printf("  code\n");
		if ((k & 0x00000040) != 0) printf("  init-data\n");
		if ((k & 0x00000080) != 0) printf("  uninit-data\n");
		if ((k & 0x10000000) != 0) printf("  sharable\n");
		if ((k & 0x20000000) != 0) printf("  AR:Exec\n");
		if ((k & 0x40000000) != 0) printf("  AR:Read\n");
		if ((k & 0x80000000) != 0) printf("  AR:Write\n");
	}

	// load
	im = malloc(imgsz);
	for (j = 0; j < imgsz; j++)
		im[j] = 0;
	for (j = 0; j < secs; j++) {
		k = get32b(&b[i + j * 40 + 20]);	// fileOfs
		if (k == 0) continue;
		m = get32b(&b[i + j * 40 + 12]);	// addr
		n = get32b(&b[i + j * 40 +  8]);	// size
		if (n > get32b(&b[i + j * 40 + 16]))
			n = get32b(&b[i + j * 40 + 16]);	// fileSiz
		for (l = 0; l < n; l++)
			im[m + l] = b[k + l];
	}

	// import
	for (j = 0; ; j++) {
		k = get32b(&im[idat + j * 20 + 12]);
		if (k == 0) break;
		printf("\nimport \"%s\":\n", &im[k]);
		k = get32b(&im[idat + j * 20 + 16]);
		for (l = 0; ; l++) {
			m = get32b(&im[k + l * 4]);
			if (m == 0) break;
			if ((m & 0x80000000) != 0)
				printf("  0x%08x: ordinal:0x%x\n", bas + k + l * 4, m & 0x7fffffff);
			else
				printf("  0x%08x: hint:0x%04x name:\"%s\"\n", bas + k + l * 4, im[m] | im[m + 1] << 8, &im[m + 2]);
		}
	}

	return 0;
}

