// jck0 for x86

#include "jck0.h"
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>

#define EnableEc	1
#define DisableEc	0

#define EAX		0
#define ECX		1
#define EDX		2

#define VARSIZ		256
#define LABSIZ		256

#define STKARGSIZ			128

typedef struct Var_ {
	const unsigned char *nam, *typ;
	int ofs;
} Var;

typedef struct Work_ {
	unsigned char *dst;
	Var var[VARSIZ], lab[LABSIZ];
	int ofs1, ec;
} Work;

static int putCodeSub(unsigned char c)
{
	if ('0' <= c && c <= '9') return c - '0';
	if ('A' <= c && c <= 'F') return c - ('A' - 10);
	if ('a' <= c && c <= 'f') return c - ('a' - 10);
	return -1;
}

static int putCode(unsigned char *q, const char *f, ...)
{
	va_list ap;
	int i = 0, j, k = 0;
	va_start(ap, f);
	while (f[i] != 0) {
		if (f[i] == ' ' || f[i] == '\t') {
			i++;
			continue;
		}
		if (putCodeSub(f[i]) >= 0) {
			q[k++] = putCodeSub(f[i]) * 16 + putCodeSub(f[i + 1]);
		} else if (f[i] == '%') {
			if (f[i + 1] == '4' || f[i + 1] == 'r') {
				j = va_arg(ap, int);
				if (f[i + 1] == 'r')
					j -= (int) &q[k + 4];
				q[k++] =  j        & 0xff;
				q[k++] = (j >>  8) & 0xff;
				q[k++] = (j >> 16) & 0xff;
				q[k++] = (j >> 24) & 0xff;
			} else if (f[i + 1] == '1') {
				j = va_arg(ap, int);
				q[k++] =  j;
			} else
				break;
		} else
			break;
		i += 2;
	}
	va_end(ap);
	return k;
}

static int nextToken(const unsigned char *src, int i)
{
	if (src[i] != 0) {
		do {
			i++;
		} while (src[i] != ' ' && src[i] != 0);
		if (src[i] == ' ') i++;
	}
	return i;
}

static int ssc(const unsigned char *s, const unsigned char *t)
{
	const unsigned char *u = strchr(t, ' ');
	if (u == 0) return 1;
	return strncmp(s, t, u - t + 1);
}

static int mssc(const unsigned char *s, const unsigned char *t)
{
	int i;
	for (i = 0; ; i++) {
		const unsigned char *u = strchr(t, ' ');
		if (u == 0) return -1;
		if (strncmp(s, t, u - t + 1) == 0) return i;
		t = u + 1;
	}
}

static int searchVar(Work *w, Var *v, const unsigned char *nam, int ecf)
{
	int i;
	for (i = 0; v[i].ofs != -2; i++) {
		if (ssc(v[i].nam, nam) == 0)
			return i;
	}
	if (ecf == 0) return -1;
	if (w->ec == 0) w->ec = jck0_ERR_UNDEF;
	return 0;
}

static int getNumber(Work *w, unsigned char *q, const unsigned char *p, int reg)
{
	int i;
	if ('0' <= *p && *p <= '9') return putCode(q, "%1 %4", 0xb8 + reg, strtol(p, 0, 0));	// MOV EAX,const.
	i = searchVar(w, w->var, p, EnableEc);
	if (i < 0) i = 0;
	return putCode(q, "8b %1 %4", 0x85 + reg * 8, w->var[i].ofs - w->ofs1);
}

static int getOfs(Work *w, const unsigned char *p)
{
	int i;
	if ('0' <= *p && *p <= '9') return strtol(p, 0, 0) - (int) w->dst;
	i = searchVar(w, w->lab, p, 0);
	if (i < 0 && w->ofs1 > 0 && w->ec == 0) w->ec = jck0_ERR_UNDEF;
	if (i < 0) i = 0;
	return w->lab[i].ofs;
}

static int testConst(const unsigned char *p)
{
	if ('0' <= *p && *p <= '9') return 1;
	return 0;
}

int jck0(const unsigned char *src, unsigned char *dst, int siz1, int *pc0, int *pc1, int *pc2)
{
	int pc, qc, pcf[8], i, j, ph, ofs, pcfl;
	Work w;
	w.lab[0].ofs = -2;
	w.dst = dst;
	w.ofs1 = 0;
	w.ec = 0;
	*pc2 = 0;
	for (ph = 0; ph < 2; ph++) {
		w.var[0].ofs = -2;
		pc = qc = ofs = 0;
		for (;;) {
			while (src[pc] == ' ' || src[pc] == '\n' || src[pc] == '\t') pc++;
			if (src[pc] == '/' && src[pc + 1] == '/') {
				while (src[pc] != 0 && src[pc] != '\n') pc++;
				continue;
			}
			if (w.ec != 0) return w.ec;
			*pc0 = *pc1 = pc;
			if (qc + 64 >= siz1) return jck0_ERR_OUTPUT;	// obt@̕s.
			pcf[0] = pc;
			pcfl = -1;
			for (i = 0; i < 8; i++) {
				if (i > 0) pcf[i] = nextToken(src, pcf[i - 1]);
				if (pcfl < 0 && ssc(&src[pcf[i]], "; ") == 0) pcfl = i;
			}
			if (src[pc] == 0) break;
			if (pcfl == 0) {
			} else if (ssc(&src[pc], "$startup ") == 0 && pcfl == 1) {
				if (w.ofs1 + STKARGSIZ >= 4096) return -999;	// [ToDo] allocaΉ.
				qc += putCode(&dst[qc], "55 89 e5 81 ec %4 ", w.ofs1 + STKARGSIZ);
			} else if (ssc(&src[pc], "$return ") == 0 && pcfl <= 2) {
				if (pcfl == 2)
					qc += getNumber(&w, &dst[qc], &src[pcf[1]], EAX);
				qc += putCode(&dst[qc], "c9 c3");
			} else if (mssc(&src[pc], "$local $temp ") >= 0) {
				if (ssc(&src[pcf[1]], "$int32 ") == 0) {
					for (pc = pcf[2]; src[pc] != 0 && ssc(&src[pc], "; ") != 0; pc = nextToken(src, pc)) {
						if (searchVar(&w, w.var, &src[pc], DisableEc) >= 0) return jck0_ERR_REDEF;	// ϐ̓d`.
						for (i = 0; w.var[i].ofs >= 0; i++);
						if (i == VARSIZ - 1) return jck0_ERR_OVRDEF;	// ϐ̓o^I[o[.
						if (w.var[i].ofs == -2) w.var[i + 1].ofs = -2;
						w.var[i].nam = &src[pc];
						w.var[i].typ = &src[pcf[1]];
						w.var[i].ofs = ofs;
						ofs += 4;
					}
					pcfl = -1;
				}
			} else if (ssc(&src[pc], "$label ") == 0 && pcfl == 2) {
				i = searchVar(&w, w.lab, &src[pcf[1]], DisableEc);
				if (w.ofs1 == 0) {
					if (i >= 0) return jck0_ERR_REDEF;	// d`.
					for (i = 0; w.lab[i].ofs >= 0; i++);
					if (i == LABSIZ - 1) return jck0_ERR_OVRDEF;	// o^I[o[.
					if (w.lab[i].ofs == -2) w.lab[i + 1].ofs = -2;
					w.lab[i].nam = &src[pcf[1]];
					w.lab[i].typ = &src[pc];
				}
				w.lab[i].ofs = qc;
			} else if (ssc(&src[pcf[1]], "= ") == 0) {
				if (src[pc] == '$' && src[pc + 1] == '#') {
					j = strtol(&src[pc + 2], 0, 0) * 4 - (w.ofs1 + STKARGSIZ);
				} else {
					i = searchVar(&w, w.var, &src[pc], EnableEc);
					j = w.var[i].ofs - w.ofs1;
				}
				qc += getNumber(&w, &dst[qc], &src[pcf[2]], EAX);
				if (pcfl > 3) {	// PȊO.
					// [ToDo] PZq̃`FbN͂ł.
					if (pcfl == 5) {
						i = mssc(&src[pcf[3]], "+ | ? ? & - ^ ? * / % ");
						if (i < 0 || ssc(&src[pcf[3]], "? ") == 0) return jck0_ERR_SYNTAX;
						qc += getNumber(&w, &dst[qc], &src[pcf[4]], ECX);
						if (i <= 7)
							qc += putCode(&dst[qc], "%1 %1", i * 8 + 3, ECX + EAX * 8 + 0xc0);	// op EAX,ECX : ADD(03), OR(0b), AND(23), SUB(2b), XOR(33).
						if (i == 8)
							qc += putCode(&dst[qc], "0f af %1", ECX + EAX * 8 + 0xc0);	// IMUL EAX,ECX.
						if (i == 9)
							qc += putCode(&dst[qc], "99 f7 f9");	// CDQ / IDIV ECX.
						if (i == 10)
							qc += putCode(&dst[qc], "99 f7 f9 89 d0");	// CDQ / IDIV ECX / MOV EAX,EDX.
					} else
						return jck0_ERR_SYNTAX;
				}
				qc += putCode(&dst[qc], "89 85 %4", j);	// MOV [EBP-...],EAX.
			} else if (ssc(&src[pc], "$if ") == 0 && ssc(&src[pcf[4]], "$goto ") == 0 && pcfl == 6) {
				qc += getNumber(&w, &dst[qc], &src[pcf[1]], EAX);
				if (testConst(&src[pcf[3]]) == 0) {
					qc += getNumber(&w, &dst[qc], &src[pcf[3]], EDX);
					qc += putCode(&dst[qc], "39 %1", EAX + EDX * 8 + 0xc0);	// CMP EAX,EDX.
				} else
					qc += putCode(&dst[qc], "81 %1 %4", EAX + 0xf8, strtol(&src[pcf[3]], 0, 0));	// CMP EAX,imm32.
				i = getOfs(&w, &src[pcf[5]]);
				j = mssc(&src[pcf[2]], "== != < >= <= > ");
				if (j < 0) return jck0_ERR_SYNTAX;
				qc += putCode(&dst[qc], "0f %1 %r", 0x80 + putCodeSub("45cdef"[j]), &dst[i]);
			} else if ((j = mssc(&src[pc], "$call $goto ")) >= 0 && pcfl == 2) {
				i = getOfs(&w, &src[pcf[1]]);
				qc += putCode(&dst[qc], "%1 %r", 0xe8 + j, &dst[i]);
			} else if (ssc(&src[pc], "$db ") == 0) {
				for (pc = pcf[1]; src[pc] != 0 && ssc(&src[pc], "; ") != 0; pc = nextToken(src, pc))
					dst[qc++] = strtol(&src[pc], 0, 16);
				pcfl = -1;
			} else if (src[pc] == '$') {
				for (pc = pcf[1]; src[pc] != 0 && ssc(&src[pc], "; ") != 0; pc = nextToken(src, pc)); // Ƃ肠S.
				pcfl = -1;
			} else
				return jck0_ERR_SYNTAX;
			if (pcfl >= 0) pc = pcf[pcfl + 1];
		}
		w.ofs1 = ofs;
		while ((w.ofs1 & 15) != 8)
			w.ofs1++;
	}
	return qc;
}

#if (defined(_WIN32))

#include <windows.h>

void *mallocRWX(int siz)
{
	return VirtualAlloc(0, siz, MEM_COMMIT, PAGE_EXECUTE_READWRITE);
}

#elif (defined(__APPLE__))

#include <mach/mach.h>

void *mallocRWX(int siz)
{
	void *p = malloc(siz);
	vm_protect(mach_task_self(), (vm_address_t) p, siz, FALSE, VM_PROT_READ | VM_PROT_WRITE | VM_PROT_EXECUTE);
	return p;
}

#elif (defined(__linux__))

#include <sys/mman.h>

void *mallocRWX(int siz)
{
	void *p = malloc(siz);
	int p0 = (int) p & ~0xfff, p1 = (((int) p) + siz1 + 0xfff) & ~0xfff;
	mprotect((void *) p0, p1 - p0, PROT_READ | PROT_WRITE | PROT_EXEC);
	return p;
}

#endif
