// jck0 for x86

#include "jck0.h"
#include "jcklib.h"
#include <string.h>
#include <stdio.h>
#include <stdlib.h>

#define EnableEc	1
#define DisableEc	0

#define EAX		0
#define ECX		1
#define EDX		2
#define EBX		3
#define ESP		4
#define EBP		5
#define ESI		6
#define EDI		7

#define VARSIZ		1024
#define LABSIZ		512
#define JMPSIZ		1024

#define STKARGSIZ			128

typedef struct Var_ {
	const unsigned char *nam, *typ;
	int ofs[2];
} Var;

typedef struct Work_ {
	const unsigned char *src;
	unsigned char *dst;
	Var var[VARSIZ], lab[LABSIZ];
	int jmp[JMPSIZ];
	int ph, ofs1, ec, pc1, opt[8], stksiz, var1, lab1;
	const unsigned char *regCache[8];
	int regUse[8];
} Work;

static int searchVar(Work *w, Var *v, int v1, const unsigned char *nam, int ecf)
{
	int i;
	for (i = v1 - 1; i >= 1; i--) {
		if (ssc(v[i].nam, nam) == 0)
			return i;
	}
	if (v == w->var && nam[0] == '$' && nam[1] == '#') {
		v[0].nam = nam;
		v[0].typ = "$int32 ";
		v[0].ofs[0] = strtol(&nam[2], 0, 0) * 4;
		if (nam[2] == 'i')
			v[0].ofs[0] = EAX - 8;
		return 0;
	}
	if (ecf == 0) return -1;
	if (v == w->lab && w->ph == 0) return 0;
	if (w->ec == 0) {
		w->ec = jck0_ERR_UNDEF;
		w->pc1 = nam - w->src;
	}
	return 0;
}

static int putCodeSub(unsigned char c)
{
	if ('0' <= c && c <= '9') return c - '0';
	if ('A' <= c && c <= 'F') return c - ('A' - 10);
	if ('a' <= c && c <= 'f') return c - ('a' - 10);
	return -1;
}

static inline int put32b(unsigned char *q, int k, int j)
{
	q[k++] =  j        & 0xff;
	q[k++] = (j >>  8) & 0xff;
	q[k++] = (j >> 16) & 0xff;
	q[k++] = (j >> 24) & 0xff;
	return k;
}

static int putCode(Work *w, unsigned char *q, const char *f, ...)
{
	va_list ap;
	int i = 0, j, k = 0, n;
	va_start(ap, f);
	while (f[i] != 0) {
		if (f[i] == ' ' || f[i] == '\t') {
			i++;
			continue;
		}
		if (putCodeSub(f[i]) >= 0) {
			q[k++] = putCodeSub(f[i]) * 16 + putCodeSub(f[i + 1]);
		} else if (f[i] == '%') {
			if (f[i + 1] == '4' || f[i + 1] == 'r') {
				j = va_arg(ap, int);
				if (f[i + 1] == 'r')
					j -= (int) &q[k + 4];
				k = put32b(q, k, j);
			} else if (f[i + 1] == '1') {
				j = va_arg(ap, int);
				q[k++] = j & 0xff;
			} else if (f[i + 1] == 'v') {
				const unsigned char *v = va_arg(ap, const unsigned char *);
				n = va_arg(ap, int) << 3;
				j = w->var[searchVar(w, w->var, w->var1, v, EnableEc)].ofs[0];
				if (j < 0)
					q[k++] = 0xc0 | n | (j + 8);
				else {
					if (w->opt[0] == 0) {
						j -= w->stksiz;
						if (w->ph == 0 || !(-128 <= j && j <= 127)) {
							q[k++] = 0x85 | n;
							k = put32b(q, k, j);
						} else {
							q[k++] = 0x45 | n;
							q[k++] = j & 0xff;
						}
					} else {
						if (w->ph == 0 || !(-128 <= j && j <= 127)) {
							q[k++] = 0x84 | n;
							q[k++] = 0x24;
							k = put32b(q, k, j);
						} else {
							q[k++] = 0x44 | n;
							q[k++] = 0x24;
							q[k++] = j & 0xff;
						}
					}
				}
			} else if (f[i + 1] == ':') {
				j = va_arg(ap, int);
				if (w->ph > 0 && j == 0)
					i += 2;
			} else
				break;
		} else
			break;
		i += 2;
	}
	va_end(ap);
	return k;
}

static int loadReg(Work *w, unsigned char *q, const unsigned char *p, int reg)
{
	int i;
	if (ssc(w->regCache[reg], p) == 0) return 0;
	w->regCache[reg] = p;
	if (testConstInt(p) != 0) return putCode(w, q, "%1 %4", 0xb8 + reg, getConstInt(p));	// MOV reg,const.
	i = searchVar(w, w->var, w->var1, p, EnableEc);
	if (w->var[i].ofs[0] < 0 && w->var[i].ofs[0] + 8 == reg) return 0;
	return putCode(w, q, "8b %v", p, reg);
}

static int getOfs(Work *w, const unsigned char *p)
{
	int i;
	if (testConstInt(p) != 0) return getConstInt(p) - (int) w->dst;
	i = searchVar(w, w->lab, w->lab1, p, EnableEc);
	if (w->ph == 2) return w->lab[i].ofs[1];
	return w->lab[i].ofs[0];
}

static int saveReg(Work *w, int qc, int r, char *f)
{
	static char *t[3] = { "#$31 ", "#$29 ", "#$30 " };	// STKARGSIZ=128z.
	int i = w->regUse[r];
	if (i >= 0 && *f == 0) {
		qc += putCode(w, &w->dst[qc], "89 %v", t[r], r);
		w->var[w->regUse[EAX]].ofs[0] = strtol(t[r] + 2, 0, 0) * 4;
		*f = 1;
	}
	return qc;
}

static int restReg(Work *w, int qc, int r)
{
	int i = w->regUse[r];
	qc += loadReg(w, &w->dst[qc], w->var[i].nam, r);
	w->var[i].ofs[0] = r - 8;
	return qc;
}

static int op8183(Work *w, int qc, int modRM, int j)
{
	unsigned char c = modRM & 0xf8;
	if (j == 0 && (c == 0xc0 || c == 0xc8 || c == 0xe8 || c == 0xf0)) return qc; // ADD(0)/OR(1)/SUB(5)/XOR(6).
	if (j == -1 && c == 0xe0) return qc; // AND(4).
	if ((j == 1 && c == 0xc0) || (j == -1 && c == 0xe8)) { w->dst[qc] = 0x40 + (modRM & 7); return qc + 1; }	// INC reg.
	if ((j == 1 && c == 0xe8) || (j == -1 && c == 0xc0)) { w->dst[qc] = 0x48 + (modRM & 7); return qc + 1; }	// DEC reg.
	if ((modRM & 0xc7) == 0xc0 && !(-128 <= j && j <= 127)) return qc + putCode(w, &w->dst[qc], "%1 %4", (modRM & 0x38) + 5, j);	// EAXɑ΂ʌ`.
	if (-128 <= j && j <= 127)
		qc += putCode(w, &w->dst[qc], "83 %1 %1", modRM, j);
	else
		qc += putCode(w, &w->dst[qc], "81 %1 %4", modRM, j);
	return qc;
}

int jck0(const unsigned char *src, unsigned char *dst, int siz1, int *pc0, int *pc1, int *pc2)
{
	int pc, qc, pcf[8], i, j, k, ofs, jmp, pcfl, r0 = 0;
	char regUse1[8];
	Work w;
	*pc2 = 0;
	w.src = src;
	w.dst = dst;
	w.ofs1 = 0;
	w.ec = 0;
	w.opt[0] = 0;
	w.opt[1] = 1;
	w.regUse[EBP] = w.regUse[ECX] = w.regUse[EDX] = w.regUse[EAX] = w.regUse[ESP] = -2;	// ftHgł͊蓖ċ֎~.
	w.regUse[EBX] = w.regUse[ESI] = w.regUse[EDI] = -1;	// 蓖ċ.
	for (i = 0; i < 8; i++)
		regUse1[i] = 0;
	for (w.ph = 0; w.ph < 3; w.ph++) {
		w.var1 = w.lab1 = 1;
		pc = qc = ofs = jmp = 0;
		for (i = 0; i < 8; i++) {
			w.regCache[i] = "; ";
			if (w.regUse[i] >= 0)
				w.regUse[i] = -1;
		}
		for (;;) {
			while (src[pc] == ' ' || src[pc] == '\n' || src[pc] == '\t') pc++;
			if (src[pc] == '/' && src[pc + 1] == '/') {
				while (src[pc] != 0 && src[pc] != '\n') pc++;
				continue;
			}
			if (w.ec != 0) {
				*pc1 = w.pc1;
				return w.ec;
			}
			*pc0 = pc;
			if (qc + 64 >= siz1) {
				*pc1 = pc;
				return jck0_ERR_OUTPUT;	// obt@̕s.
			}
			pcf[0] = pc;
			pcfl = -1;
			for (i = 0; i < 8; i++) {
				if (i > 0) pcf[i] = nextToken(src, pcf[i - 1]);
				if (pcfl < 0 && ssc(&src[pcf[i]], "; ") == 0) pcfl = i;
			}
			if (src[pc] == 0) break;
			if (pcfl == 0) {
			} else if (ssc(&src[pc], "$option ") == 0 && pcfl == 3) {
				i = strtol(&src[pcf[1]], 0, 0);
				if (0 <= i && i <= 7)
					w.opt[i] = strtol(&src[pcf[2]], 0, 0);
				if (i == 0 && w.opt[i] != 0) w.regUse[EBP] = -1;
				if (i == 0 && w.opt[i] == 0) w.regUse[EBP] = -2;
				if (i == 1 && w.opt[i] == 1) { w.regUse[EAX] = -2; w.regUse[ECX] = -2; w.regUse[EDX] = -2; } // ZϒVtgg.
				if (i == 1 && w.opt[i] == 2) { w.regUse[EAX] = -2; w.regUse[ECX] = -2; w.regUse[EDX] = -1; } // ϒVtg͎gAZ͒xĂ.
				if (i == 1 && w.opt[i] == 3) { w.regUse[EAX] = -2; w.regUse[ECX] = -1; w.regUse[EDX] = -1; } // ϒVtgxĂ.
				if (i == 1 && w.opt[i] >= 4) { w.regUse[EAX] = -1; w.regUse[ECX] = -1; w.regUse[EDX] = -1; } // ANZXxĂ. ֐̖߂l󂯎Ȃ.
			} else if (ssc(&src[pc], "$startup ") == 0 && pcfl == 1) {
				w.stksiz = w.ofs1 + STKARGSIZ;
				if (w.stksiz >= 4096) return -999;	// [ToDo] allocaΉ.
				if (w.opt[0] == 0)
					qc += putCode(&w, &dst[qc], "%:53 %:56 %:57 55 89 e5 81 ec %4", regUse1[EBX], regUse1[ESI], regUse1[EDI], w.stksiz);
				else
					qc += putCode(&w, &dst[qc], "%:53 %:56 %:57 %:55 81 ec %4", regUse1[EBX], regUse1[ESI], regUse1[EDI], regUse1[EBP], w.stksiz);
			} else if (ssc(&src[pc], "$return ") == 0 && pcfl <= 2) {
				if (pcfl == 2)
					qc += loadReg(&w, &dst[qc], &src[pcf[1]], EAX);
				if (w.opt[0] == 0)
					qc += putCode(&w, &dst[qc], "c9 %:5f %:5e %:5b c3", regUse1[EDI], regUse1[ESI], regUse1[EBX]);
				else
					qc += putCode(&w, &dst[qc], "81 c4 %4 %:5d %:5f %:5e %:5b c3", w.stksiz, regUse1[EBP], regUse1[EDI], regUse1[ESI], regUse1[EBX]);
			} else if (mssc(&src[pc], "$local $temp ") >= 0) {
				if (ssc(&src[pcf[1]], "$int32 ") == 0) {
					for (pc = pcf[2]; src[pc] != 0 && ssc(&src[pc], "; ") != 0; pc = nextToken(src, pc)) {
						if (w.var1 == VARSIZ) {
							*pc1 = pc;
							return jck0_ERR_OVRDEF;	// ϐ̓o^I[o[.
						}
						w.var[w.var1].nam = &src[pc];
						w.var[w.var1].typ = &src[pcf[1]];
						if (w.ph == 0) {
							w.var[w.var1].ofs[0] = ofs + STKARGSIZ;
							ofs += 4;
						} else {
							if (w.var[w.var1].ofs[0] >= 0) {
								w.var[w.var1].ofs[0] = ofs + STKARGSIZ;
								ofs += 4;
							}
						}
						w.var1++;
					}
					pcfl = -1;
				}
			} else if (ssc(&src[pc], "$label ") == 0 && pcfl == 2) {
				i = searchVar(&w, w.lab, w.lab1, &src[pcf[1]], DisableEc);
				if (!(i >= 0 && ssc(w.lab[i].typ, "$reserve ") == 0)) {
					if (w.lab1 == LABSIZ) {
						*pc1 = pcf[1];
						return jck0_ERR_OVRDEF;	// o^I[o[.
					}
					i = w.lab1++;
					w.lab[i].nam = &src[pcf[1]];
				}
				w.lab[i].typ = &src[pc];
				if (w.ph < 2) {
					w.lab[i].ofs[w.ph] = qc;
				} else {
					if (w.lab[i].ofs[1] != qc) {
						*pc1 = pcf[1];
						return jck0_ERR_INTERNAL;	// G[.
					}
				}
				for (i = 0; i < 8; i++)
					w.regCache[i] = "; ";
			} else if (ssc(&src[pc], "$reserve ") == 0) {
				for (pc = pcf[1]; src[pc] != 0 && ssc(&src[pc], "; ") != 0; pc = nextToken(src, pc)) {
					if (w.lab1 == LABSIZ) {
						*pc1 = pc;
						return jck0_ERR_OVRDEF;	// o^I[o[.
					}
					w.lab[w.lab1].nam = &src[pc];
					w.lab[w.lab1].typ = "$reserve ";
					if (w.ph == 0)
						w.lab[w.lab1].ofs[0] = -1;
					w.lab1++;
				}
				pcfl = -1;
			} else if (ssc(&src[pc], "$dev ") == 0) {
				for (pc = pcf[1]; src[pc] != 0 && ssc(&src[pc], "; ") != 0; pc = nextToken(src, pc)) {
					i = searchVar(&w, w.var, w.var1, &src[pc], EnableEc);
					w.var[i].nam = "; ";
					if (w.var[i].ofs[0] < 0)
						w.regUse[w.var[i].ofs[0] + 8] = -1;
					for (i = 0; i < 8; i++) {
						if (ssc(w.regCache[i], &src[pc]) == 0)
							w.regCache[i] = "; ";
					}
				}
				pcfl = -1;
			} else if (ssc(&src[pc], "$del ") == 0) {
				for (pc = pcf[1]; src[pc] != 0 && ssc(&src[pc], "; ") != 0; pc = nextToken(src, pc))
					w.lab[searchVar(&w, w.lab, w.lab1, &src[pc], EnableEc)].nam = "; ";
				pcfl = -1;
			} else if (ssc(&src[pc], "* ") == 0 && ssc(&src[pcf[2]], "= ") == 0 && pcfl == 4) {
				char savEAX = 0, savEDX = 0;
				int r0 = EAX, r1 = EDX;
				qc = saveReg(&w, qc, EAX, &savEAX);
				qc = saveReg(&w, qc, EDX, &savEDX);
				qc += loadReg(&w, &dst[qc], &src[pcf[1]], r1);
				qc += loadReg(&w, &dst[qc], &src[pcf[3]], r0);
				qc += putCode(&w, &dst[qc], "89 %1 00", r1 | r0 << 3 | 0x40);
				if (savEAX != 0) qc = restReg(&w, qc, EAX);
				if (savEDX != 0) qc = restReg(&w, qc, EDX);
				for (i = 0; i < 8; i++)
					w.regCache[i] = "; ";
			} else if (ssc(&src[pcf[1]], "= ") == 0) {
				char savEAX = 0, savECX = 0, savEDX = 0, letMem = 1;
				i = searchVar(&w, w.var, w.var1, &src[pc], EnableEc);
				if (w.var[i].ofs[0] < 0) {
					r0 = w.var[i].ofs[0] + 8;
					letMem = 0;
				}
				if (pcfl == 3) {	// P.
					if (testConstInt(&src[pcf[2]]) == 0) {
						i = searchVar(&w, w.var, w.var1, &src[pcf[2]], EnableEc);
						if (w.var[i].ofs[0] < 0) {
							qc += putCode(&w, &dst[qc], "89 %v", &src[pc], w.var[i].ofs[0] + 8);	// r/m = r. (\[XWX^).
							letMem = 0;
						} else {
							if (letMem != 0) {
								qc = saveReg(&w, qc, EAX, &savEAX);
								r0 = EAX;
							}
							qc += loadReg(&w, &dst[qc], &src[pcf[2]], r0);	// r/m = m.
						}
					} else {	// imm.
						if (letMem != 0) {
							qc = saveReg(&w, qc, EAX, &savEAX);
							r0 = EAX;
						}
						if (getConstInt(&src[pcf[2]]) == 0)
							qc += putCode(&w, &dst[qc], "31 %1", r0 * 9 + 0xc0);	// XOR r0,r0.
						else
							qc += loadReg(&w, &dst[qc], &src[pcf[2]], r0);	// r = imm.
					}
				} else if (pcfl == 4) {	// PZq.
					if (letMem != 0) {
						qc = saveReg(&w, qc, EAX, &savEAX);
						r0 = EAX;
					}
					if (ssc(&src[pcf[2]], "& ") == 0) {
						qc += putCode(&w, &dst[qc], "8d %v", &src[pcf[3]], r0);	// LEA r0,r/m.
					} else if (ssc(&src[pcf[2]], "* ") == 0) {
						qc += loadReg(&w, &dst[qc], &src[pcf[3]], r0);
						qc += putCode(&w, &dst[qc], "8b %1 00", r0 | r0 << 3 | 0x40);
					} else {
						*pc1 = pcf[2];
						return jck0_ERR_SYNTAX;
					}
				} else if (pcfl == 5) {	// 񍀉Zq.
					i = mssc(&src[pcf[3]], "+ | ? ? & - ^ ? * / % ");
					if (i < 0 || ssc(&src[pcf[3]], "? ") == 0) {
						*pc1 = pcf[3];
						return jck0_ERR_SYNTAX;
					}
					j = getConstInt(&src[pcf[4]]);
					if (i <= 8) {
						if (letMem != 0) {
							qc = saveReg(&w, qc, EAX, &savEAX);
							r0 = EAX;
						}
						if (i <= 7) {
							qc += loadReg(&w, &dst[qc], &src[pcf[2]], r0);
							if (testConstInt(&src[pcf[4]]) == 0)
								qc += putCode(&w, &dst[qc], "%1 %v", i * 8 + 3, &src[pcf[4]], r0);	// op r0,r/m : ADD(03), OR(0b), AND(23), SUB(2b), XOR(33).
							else
								qc = op8183(&w, qc, r0 + i * 8 + 0xc0, j);	// op r0,imm
						}
						if (i == 8) {
							if (testConstInt(&src[pcf[4]]) == 0) {
								qc += loadReg(&w, &dst[qc], &src[pcf[2]], r0);
								qc += putCode(&w, &dst[qc], "0f af %v", &src[pcf[4]], r0);	// IMUL r0,r/m.
							} else
								qc += putCode(&w, &dst[qc], "69 %v %4", &src[pcf[2]], r0, j);	// IMUL r0,r/m,imm32.
						}
					}
					if (i == 9 || i == 10) {
						if (!(letMem == 0 && i ==  9 && r0 == EAX))
							qc = saveReg(&w, qc, EAX, &savEAX);
						if (!(letMem == 0 && i == 10 && r0 == EDX))
							qc = saveReg(&w, qc, EDX, &savEDX);
						if (testConstInt(&src[pcf[4]]) == 0) {
							qc += loadReg(&w, &dst[qc], &src[pcf[2]], EAX);
							qc += putCode(&w, &dst[qc], "99 f7 %v", &src[pcf[4]], 7);
						} else {
							qc = saveReg(&w, qc, ECX, &savECX);
							qc += loadReg(&w, &dst[qc], &src[pcf[4]], ECX);
							qc += loadReg(&w, &dst[qc], &src[pcf[2]], EAX);
							qc += putCode(&w, &dst[qc], "99 f7 f9");	// IDIV ECX.
						}
						w.regCache[EAX] = w.regCache[EDX] = "; ";
						if (i ==  9) j = EAX;
						if (i == 10) j = EDX;
						if (letMem != 0)
							r0 = j;
						if (letMem == 0 && r0 != j)
							qc += putCode(&w, &dst[qc], "89 %1", r0 + j * 8 + 0xc0);	// r0 = j (r/m = r).
					}
				} else {
					*pc1 = pcf[1];
					return jck0_ERR_SYNTAX;
				}
				if (letMem != 0) {
					qc += putCode(&w, &dst[qc], "89 %v", &src[pc], r0);
					w.regCache[r0] = &src[pc];
				}
				if (savEAX != 0) qc = restReg(&w, qc, EAX);
				if (savECX != 0) qc = restReg(&w, qc, ECX);
				if (savEDX != 0) qc = restReg(&w, qc, EDX);
			} else if (ssc(&src[pc], "$if ") == 0 && ssc(&src[pcf[4]], "$goto ") == 0 && pcfl == 6) {
				char savEAX = 0;
				if (jmp >= JMPSIZ) {
					*pc1 = pc;
					return jck0_ERR_OVRDEF;	// o^I[o[.
				}
				i = searchVar(&w, w.var, w.var1, &src[pcf[1]], EnableEc);
				j = getConstInt(&src[pcf[3]]);
				if (w.var[i].ofs[0] >= 0 && testConstInt(&src[pcf[3]]) != 0 && w.regUse[EAX] >= 0) {
					if (-128 <= j && j <= 127)
						qc += putCode(&w, &dst[qc], "83 %v %1", &src[pcf[1]], 7, j);	// CMP r/m,imm8.
					else
						qc += putCode(&w, &dst[qc], "81 %v %4", &src[pcf[1]], 7, j);	// CMP r/m,imm32.
				} else {
					if (w.var[i].ofs[0] < 0)
						r0 = w.var[i].ofs[0] + 8;
					else {
						qc = saveReg(&w, qc, EAX, &savEAX);
						r0 = EAX;
						qc += loadReg(&w, &dst[qc], &src[pcf[1]], r0);
					}
					if (testConstInt(&src[pcf[3]]) == 0)
						qc += putCode(&w, &dst[qc], "3b %v", &src[pcf[3]], r0);	// CMP r0,r/m.
					else
						qc = op8183(&w, qc, r0 + 0xf8, j);	// CMP r0,imm.
					if (savEAX != 0) qc = restReg(&w, qc, EAX);
				}
				i = searchVar(&w, w.lab, w.lab1, &src[pcf[5]], EnableEc);
				if (w.ec == 0 && w.ph > 0 && w.lab[i].ofs[0] < 0) { *pc1 = pcf[5]; return jck0_ERR_UNDEF; }
				j = mssc(&src[pcf[2]], "== != < >= <= > ");
				if (j < 0) {
					*pc1 = pcf[2];
					return jck0_ERR_SYNTAX;
				}
				if (w.ph == 0)
					w.jmp[jmp] = qc + 6;
				k = 128;
				if (w.ph > 0)
					k = w.lab[i].ofs[0] - w.jmp[jmp];
				if (-128 <= k && k <= 127 && k != 0)	//k==0ȂCMPȂĂ܂Ƃ낾ǁA܂ł͂ĂȂ.
					qc += putCode(&w, &dst[qc], "%1 %1", 0x70 + putCodeSub("45cdef"[j]), w.lab[i].ofs[1] - (qc + 2));
				else if (k != 0)
					qc += putCode(&w, &dst[qc], "0f %1 %r", 0x80 + putCodeSub("45cdef"[j]), &dst[w.lab[i].ofs[1]]);
				jmp++;
			} else if (ssc(&src[pc], "$goto ") == 0 && pcfl == 2) {
				if (jmp >= JMPSIZ) {
					*pc1 = pc;
					return jck0_ERR_OVRDEF;	// o^I[o[.
				}
				i = searchVar(&w, w.lab, w.lab1, &src[pcf[1]], EnableEc);
				if (w.ec == 0 && w.ph > 0 && w.lab[i].ofs[0] < 0) { *pc1 = pcf[1]; return jck0_ERR_UNDEF; }
				if (w.ph == 0)
					w.jmp[jmp] = qc + 5;
				k = 128;
				if (w.ph > 0)
					k = w.lab[i].ofs[0] - w.jmp[jmp];
				if (-128 <= k && k <= 127 && k != 0)
					qc += putCode(&w, &dst[qc], "eb %1", (w.lab[i].ofs[1] - (qc + 2)) & 0xff);
				else if (k != 0)
					qc += putCode(&w, &dst[qc], "e9 %r", &dst[w.lab[i].ofs[1]]);
				jmp++;
			} else if (ssc(&src[pc], "$call ") == 0 && pcfl == 2) {
				char savEAX = 0, savECX = 0, savEDX = 0;
				qc = saveReg(&w, qc, EAX, &savEAX);
				qc = saveReg(&w, qc, ECX, &savECX);
				qc = saveReg(&w, qc, EDX, &savEDX);
				qc += putCode(&w, &dst[qc], "e8 %r", &dst[getOfs(&w, &src[pcf[1]])]);
				if (savEAX != 0) qc = restReg(&w, qc, EAX);
				if (savECX != 0) qc = restReg(&w, qc, ECX);
				if (savEDX != 0) qc = restReg(&w, qc, EDX);
			} else if (ssc(&src[pc], "$db ") == 0) {
				for (pc = pcf[1]; src[pc] != 0 && ssc(&src[pc], "; ") != 0; pc = nextToken(src, pc))
					dst[qc++] = strtol(&src[pc], 0, 16);
				pcfl = -1;
			} else if (ssc(&src[pc], "$reg ") == 0) {
				static int t[8] = { EBX, ESI, EDI, EBP, EDX, ECX, EAX, ESP };
				for (pc = pcf[1]; src[pc] != 0 && ssc(&src[pc], "; ") != 0; pc = nextToken(src, pc)) {
					i = searchVar(&w, w.var, w.var1, &src[pc], EnableEc);
					// [ToDo] ^sKȂskip.
					if (w.var[i].ofs[0] < 0) continue;
					for (j = 0; j < 8; j++) {
						if (w.regUse[t[j]] == -1) break;
					}
					if (j < 8) {
						w.var[i].ofs[0] = t[j] - 8;
						w.regUse[t[j]] = i;
						regUse1[t[j]] = 1;
					}
				}
				pcfl = -1;
			} else {
				*pc1 = pc;
				return jck0_ERR_SYNTAX;
			}
			if (pcfl >= 0) pc = pcf[pcfl + 1];
		}
		w.ofs1 = ofs;
		i = (24 - (regUse1[EBX] + regUse1[ESI] + regUse1[EDI]) * 4) & 15;
		while ((w.ofs1 & 15) != i)
			w.ofs1++;
	}
	return qc;
}

#if (defined(_WIN32))

#include <windows.h>

void *mallocRWX(int siz)
{
	return VirtualAlloc(0, siz, MEM_COMMIT, PAGE_EXECUTE_READWRITE);
}

#elif (defined(__APPLE__))

#include <mach/mach.h>

void *mallocRWX(int siz)
{
	void *p = malloc(siz);
	vm_protect(mach_task_self(), (vm_address_t) p, siz, FALSE, VM_PROT_READ | VM_PROT_WRITE | VM_PROT_EXECUTE);
	return p;
}

#elif (defined(__linux__))

#include <sys/mman.h>

void *mallocRWX(int siz)
{
	void *p = malloc(siz);
	int p0 = (int) p & ~0xfff, p1 = (((int) p) + siz1 + 0xfff) & ~0xfff;
	mprotect((void *) p0, p1 - p0, PROT_READ | PROT_WRITE | PROT_EXEC);
	return p;
}

#endif
